import asyncssh

from actappliance.connections.aioconnection import AIOConnection
from actappliance.connections.ssh import ssh_out


class AIOSsh(AIOConnection):
    """
    self._conn holds an ssh connection in this class
    """
    def __init__(self, *args, **kwargs):
        super(AIOSsh, self).__init__(*args, **kwargs)

    def connect(self, *args, **kwargs):
        self._loop.run_until_complete(self._connect(*args, **kwargs))

    async def _connect(self, *args, **kwargs):
        kwargs.setdefault('port', 26)
        kwargs.setdefault('client_factory', None)
        self._conn, _ = await asyncssh.create_connection(*args, host=self.system, loop=self._loop,
                                                         **{**self.connect_kwargs, **kwargs})

    async def _disconnect(self):
        await self._conn.__aexit__()

    async def _single_conn_request(self, *args, **kwargs):
        async with asyncssh.connect(self.system, **self.connect_kwargs) as conn:
            await self._conn_request(*args, conn=conn, **kwargs)

    async def _conn_request(self, act_cmd_obj, *, conn, check: bool=True):
        # Can't use ssh_kwargs because run's input is 'cmd' not 'command'
        result = await conn.run(act_cmd_obj.input.ssh_command, check=True)

        ssh_o = ssh_out(
            result.stdout.rstrip('\n').split('\n'),
            result.stderr.rstrip('\n').split('\n'),
            result.exit_status
        )
        act_cmd_obj.output.sshout = ssh_o

    def scp(self, srcpaths, path=b'.', **kwargs):
        """Copy file(s) to a destination path using scp

        :param srcpaths: See asyncssh.scp
        :param path: A str only input will be used as a path instead of a connection with a default path of b'.'
            Note: This is not the same as asyncssh.scp as it's dstpath assumes a str input is
        :param kwargs: See asyncssh.scp
        """
        # Comparison from asyncssh.scp's _parse_path
        if not isinstance(path, (str, bytes)):
            # The connection will be added by the class's connection and shouldn't be passed through this method
            raise ValueError(f"Path input should be destination path only.")
        dstpath = (self._conn, path)
        self._loop.run_until_complete(asyncssh.scp(srcpaths, dstpath, **kwargs))



import asyncio
from abc import ABC, abstractmethod

from actappliance.connections import CALL_TIMEOUT
from actappliance.models import ActCmd


class AIOConnection(ABC):
    def __init__(self, system, connect_kwargs: dict=None, *, call_kwargs: dict=None, loop=None,
                 call_timeout=CALL_TIMEOUT):
        self.system = system
        self.connect_kwargs = connect_kwargs if connect_kwargs is not None else {}
        self.call_kwargs = call_kwargs if call_kwargs is not None else {}
        self._loop = loop or asyncio.get_event_loop()
        self.timeout = call_timeout
        self._conn = None

    def connect(self, *args, **kwargs):
        self._loop.run_until_complete(self._connect(*args, **kwargs))

    @abstractmethod
    async def _connect(self, *args, **kwargs):
        pass

    def disconnect(self):
        self._loop.run_until_complete(self._disconnect())

    @abstractmethod
    async def _disconnect(self):
        pass

    async def _call(self, act_cmd_obj: ActCmd):
        await self._request(act_cmd_obj)

    async def _request(self, act_cmd_obj: ActCmd, *, check: bool=True):
        """Make the request with the same conn from connect or an ephemeral connection if the user hasn't connected"""
        if not self._conn:
            return await self._single_conn_request(act_cmd_obj, check=check)
        return await self._conn_request(act_cmd_obj, conn=self._conn, check=check)

    @abstractmethod
    async def _single_conn_request(self, act_cmd_obj, *, check: bool=True):
        """Create an ephemeral connection and then make a request with it"""
        pass

    @abstractmethod
    async def _conn_request(self, act_cmd_obj, *, conn, check: bool=True):
        """Make a request with a provided connection"""
        pass

    async def _timeout_call(self, *args, **kwargs):
        fut = self._call(*args, **kwargs)
        # Consider the speed benefits of asyncio_timeout.timeout vs added dep
        # Warning: this behaves differently <3.7 because wait_for wouldn't wait for future to cancel before Error
        result = await asyncio.wait_for(fut, self.timeout, loop=self._loop)
        return result

    def cmd(self, *args, **update_cmds):
        return self._loop.run_until_complete(self._cmd(*args, **update_cmds))

    async def _cmd(self, operation, **update_cmds):
        ac = ActCmd(conn_object=self)
        ac.cmd_input(operation, **update_cmds)
        await self._timeout_call(ac)
        return ac.response



import aiohttp

from actappliance.connections.aioconnection import AIOConnection

AIO_TIMEOUT = aiohttp.ClientTimeout(total=None, connect=121, sock_connect=121, sock_read=241)


class AIORest(AIOConnection):
    """
    self._conn holds an aiohttp session in this class
    """
    async def _connect(self, *args, **kwargs):
        # These kwargs are kwo so no arg handling required
        kwargs.setdefault('timeout', AIO_TIMEOUT)
        kwargs.setdefault('loop', self._loop)
        kwargs.setdefault('headers', {'Content-Type': 'application/json'})
        self._conn = aiohttp.ClientSession(*args, **{**self.connect_kwargs, **kwargs})
        result = await self._conn.post(f"https://{self.system}/actifio/api/login", params=self.call_kwargs,
                                       verify_ssl=False)
        result.raise_for_status()
        sid_r = await result.json()
        self.sid = sid_r['sessionid']

    async def _disconnect(self):
        await self._conn.close()

    async def _single_conn_request(self, act_cmd_obj, *, check: bool=True):
        async with aiohttp.ClientSession() as session:
            session._raise_for_status = check
            return await self._conn_request(act_cmd_obj, conn=session)

    async def _conn_request(self, act_cmd_obj, *, conn, check: bool=True):
        print(act_cmd_obj.call_kwargs)
        r = await conn.request(**act_cmd_obj.call_kwargs, verify_ssl=False)
        if check:
            r.raise_for_status()
        act_cmd_obj.output.result = await r.json()
